/**
 * 
 */
package com.wieseke.cptk.output.dao;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;

import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.dao.Cophylogeny;
import com.wieseke.cptk.reconstruction.dao.EventType;
import com.wieseke.cptk.reconstruction.dao.ReconstructionChildMapping;
import com.wieseke.cptk.reconstruction.dao.ReconstructionCophylogeny;
import com.wieseke.cptk.reconstruction.dao.ReconstructionHostNode;
import com.wieseke.cptk.reconstruction.dao.ReconstructionParasiteNode;
import com.wieseke.cptk.reconstruction.dao.ReconstructionSolution;
import com.wieseke.cptk.reconstruction.util.CostCalculator;
import com.wieseke.cptk.reconstruction.util.QualityCalculator;

/**
 * @author Nicolas Wieseke
 * 
 */
public class OutputCophylogeny extends Cophylogeny {

	private Map<ReconstructionHostNode, OutputHostNode> hostRelations;
	private OutputHostNode[] hostNodes;
	private Map<EventType, ReconstructionSolution> bestSolutions;
	private List<Integer> rootMappings;
	private int rootMappingIndex;
	private ReconstructionCophylogeny reconstructionCophylogeny;

	public OutputCophylogeny(ReconstructionCophylogeny cophylogeny) {
		this.reconstructionCophylogeny = cophylogeny;
		setFilename(cophylogeny.getFilename());
		this.costs = cophylogeny.getCosts();
		this.options = cophylogeny.getOptions();
		this.rootMappings = cophylogeny.computeBestRootMappings();
		this.rootMappingIndex = 0;
		if (cophylogeny.getBestSolutions().size() != 0) {
			this.bestSolutions = cophylogeny.getBestSolutions();
		} else {
			this.bestSolutions = new HashMap<EventType, ReconstructionSolution>();
			for (ReconstructionSolution solution : cophylogeny
					.getCurrentSolutions()) {
				this.bestSolutions.put(solution.getEvents(), solution);
			}
		}

		init(cophylogeny);
	}

	private void init(ReconstructionCophylogeny reconstructionCophylogeny) {
		this.hostRelations = new HashMap<ReconstructionHostNode, OutputHostNode>();
		setHostRoot(initHostNodes(reconstructionCophylogeny.getHostNodes()[0],
				null));
		this.hostNodes = new OutputHostNode[reconstructionCophylogeny
				.getHostNodes().length];
		for (int i = 0; i < reconstructionCophylogeny.getHostNodes().length; i++) {
			this.hostNodes[i] = this.hostRelations
					.get(reconstructionCophylogeny.getHostNodes()[i]);
		}
		ReconstructionChildMapping mapping = new ReconstructionChildMapping();
		mapping.setChildMapping(getRootMapping());
		setParasiteRoot(initParasiteNodes(
				reconstructionCophylogeny.getParasiteNodes()[0], null,
				getRootMapping(), 0, mapping));
	}

	/**
	 * initializes the host nodes for the reconstruction according to the given
	 * input host nodes
	 * 
	 * @param reconstructionHostNode
	 * @param parent
	 * @return
	 */
	private OutputHostNode initHostNodes(
			ReconstructionHostNode reconstructionHostNode, OutputHostNode parent) {
		OutputHostNode node = new OutputHostNode(reconstructionHostNode);
		node.setParent(parent);

		for (ReconstructionHostNode childNode : reconstructionHostNode
				.getChildren()) {
			node.getChildren().add(initHostNodes(childNode, node));
		}

		this.hostRelations.put(reconstructionHostNode, node);
		return node;
	}

	/**
	 * initializes the parasite nodes for the reconstruction according to the
	 * given input parasite nodes data
	 * 
	 * @param inputParasiteNode
	 * @param parent
	 * @return
	 */
	private OutputParasiteNode initParasiteNodes(
			ReconstructionParasiteNode reconstructionParasiteNode,
			OutputParasiteNode parent, int hostIndex, int ownChildIndex,
			ReconstructionChildMapping mapping) {
		if (hostIndex < 0) {
			return null;
		}
		OutputParasiteNode node = new OutputParasiteNode(
				reconstructionParasiteNode, ownChildIndex);
		node.setParent(parent);
		node.setHostNodes(this.hostNodes);
		node.setHostIndex(hostIndex);
		node.setMapping(mapping);

		int childIndex = 0;
		for (ReconstructionParasiteNode childNode : reconstructionParasiteNode
				.getChildren()) {
			OutputParasiteNode child = initParasiteNodes(
					childNode,
					node,
					reconstructionParasiteNode.getReconstructionInfo()[hostIndex]
							.getReconstructions().get(0).getChildMappings()[childIndex]
							.getChildMapping(),
					childIndex,
					reconstructionParasiteNode.getReconstructionInfo()[hostIndex]
							.getReconstructions().get(0).getChildMappings()[childIndex]);
			if (child != null)
				node.getChildren().add(child);
			childIndex++;
		}

		return node;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.wieseke.cptk.common.dao.Cophylogeny#getHostRoot()
	 */

	public OutputHostNode getHostRoot() {
		return (OutputHostNode) super.getHostRoot();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.wieseke.cptk.common.dao.Cophylogeny#getParasiteRoot()
	 */

	public OutputParasiteNode getParasiteRoot() {
		return (OutputParasiteNode) super.getParasiteRoot();
	}

	/**
	 * @param label
	 * @param currentNode
	 * @return true if the label is not used by any other than the current node,
	 *         else false
	 */
	public boolean checkNodeLabel(String label, IPositionNode currentNode) {
		List<IPositionNode> nodeList = getNodeList(getHostRoot());
		nodeList.addAll(getNodeList(getParasiteRoot()));
		for (IPositionNode node : nodeList) {
			if (node.getLabel().equals(label) && !node.equals(currentNode)) {
				return false;
			}
		}
		return true;
	}

	public List<IPositionNode> getNodeList(IPositionNode node) {
		List<IPositionNode> nodeList = new ArrayList<IPositionNode>();
		nodeList.add(node);
		for (IPositionNode child : node.getChildren()) {
			nodeList.addAll(getNodeList(child));
		}
		return nodeList;
	}

	public int getRootMapping() {
		return this.rootMappings.get(this.rootMappingIndex);
	}

	public int getNextRootMapping() {
		if (this.rootMappings.size() > this.rootMappingIndex + 1) {
			this.rootMappingIndex++;
			return getRootMapping();
		}
		return -1;
	}

	public int getPreviousRootMapping() {
		if (this.rootMappingIndex > 0) {
			this.rootMappingIndex--;
			return getRootMapping();
		}
		return -1;
	}

	public boolean nextReconstruction(boolean validOnly, boolean showDialog) {
		if (validOnly) {
			while (true) {
				if (getParasiteRoot().nextReconstruction()) {
					if (!isValid()) {
						continue;
					}
					return true;
				}
				if (getNextRootMapping() != -1) {
					getParasiteRoot().initFirstReconstruction(getRootMapping());
					ReconstructionChildMapping mapping = new ReconstructionChildMapping();
					mapping.setChildMapping(getRootMapping());
					getParasiteRoot().setMapping(mapping);
					if (!isValid()) {
						continue;
					}
					return true;
				}
				break;
			}
			if (showDialog) {
				try {
					MessageDialog.openInformation(PlatformUI.getWorkbench()
							.getDisplay().getShells()[0], "Info",
							"Last solution");
				} catch (Exception e) {
				}
			}
			return false;
		} else {
			if (getParasiteRoot().nextReconstruction()) {
				return true;
			}
			if (getNextRootMapping() != -1) {
				getParasiteRoot().initFirstReconstruction(getRootMapping());
				ReconstructionChildMapping mapping = new ReconstructionChildMapping();
				mapping.setChildMapping(getRootMapping());
				getParasiteRoot().setMapping(mapping);
				return true;
			}
			if (showDialog) {
				try {
					MessageDialog.openInformation(PlatformUI.getWorkbench()
							.getDisplay().getShells()[0], "Info",
							"Last solution");
				} catch (Exception e) {
				}
			}
			return false;
		}
	}

	public boolean previousReconstruction(boolean validOnly) {
		if (getParasiteRoot().previousReconstruction()) {
			return true;
		}
		if (getPreviousRootMapping() != -1) {
			getParasiteRoot().initPreviousReconstruction(getRootMapping());
			ReconstructionChildMapping mapping = new ReconstructionChildMapping();
			mapping.setChildMapping(getRootMapping());
			getParasiteRoot().setMapping(mapping);
			return true;
		}
		return false;
	}

	public boolean chooseReconstruction(long number) {
		if (number >= 0) {
			while (previousReconstruction(false)) {
			}
			for (long i = 0; i < number && nextReconstruction(false, true); i++) {
			}
			return true;
		}
		return false;
	}

	public ReconstructionSolution getSolution() {
		EventType events = getParasiteRoot().getDerivedEvents();
		return new ReconstructionSolution(getRootMapping(),
				CostCalculator.computeTotalCosts(this.costs, events),
				this.costs,
				QualityCalculator.computeQuality(this.costs, events), events);
	}

	/**
	 * @return the bestSolutions
	 */
	public Map<EventType, ReconstructionSolution> getBestSolutions() {
		return bestSolutions;
	}

	public boolean isValid() {
		ChronologyCheck chronologyCheck = new ChronologyCheck(getHostRoot(),
				getParasiteRoot());
		return chronologyCheck.isAcyclic();
	}

	public String getRootMappingName() {
		if (getRootMapping() >= 0)
			return this.hostNodes[getRootMapping()].getLabel();
		return "";
	}

	public ReconstructionCophylogeny getReconstructionCophylogeny() {
		return this.reconstructionCophylogeny;
	}

}
