/**
 * 
 */
package com.wieseke.cptk.reconstruction.dao;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import com.wieseke.cptk.common.api.IHostNode;
import com.wieseke.cptk.common.api.INode;
import com.wieseke.cptk.common.constants.OptionsConstants;
import com.wieseke.cptk.common.util.NodeUtils;
import com.wieseke.cptk.common.util.NullLogger;
import com.wieseke.cptk.reconstruction.constants.BaseSolverConstants;

/**
 * @author Nicolas Wieseke
 * 
 */
public class StaticCostTable {

	/**
	 * Logger
	 */
	// static Logger logger = Logger.getLogger(StaticCostTable.class);
	static NullLogger logger = new NullLogger();

	public StaticCostTableItem[][] costs;
	private String hostswithOption;
	private String sortingOption;
	private String takeoffOption;
	private String timezoneOption;
	private BigDecimal hostswithCosts;
	private BigDecimal sortingCosts;

	/**
	 * initializes the static cost table
	 * 
	 * @param nodes
	 * @param hostswithOption
	 * @param sortingOption
	 * @param takeoffOption
	 * @param timezoneOption
	 * @param hostswithCosts
	 * @param sortingCosts
	 */
	public void init(IHostNode[] nodes, String hostswithOption,
			String sortingOption, String takeoffOption, String timezoneOption,
			BigDecimal hostswithCosts, BigDecimal sortingCosts) {
		logger.debug("--- start calculating static cost table ---");
		try {
			this.costs = new StaticCostTableItem[nodes.length][nodes.length];
			this.hostswithOption = hostswithOption;
			this.sortingOption = sortingOption;
			this.takeoffOption = takeoffOption;
			this.timezoneOption = timezoneOption;
			this.hostswithCosts = hostswithCosts;
			this.sortingCosts = sortingCosts;
			ExecutorService threadExecutor = Executors
					.newFixedThreadPool(BaseSolverConstants.NUMBER_OF_THREADS);
			for (IHostNode start : nodes) {
				for (IHostNode end : nodes) {
					StaticCostCalculator calculator = new StaticCostCalculator();
					calculator.init(start, end);
					threadExecutor.execute(calculator);
				}
			}
			threadExecutor.shutdown();
			while (!threadExecutor
					.awaitTermination(1000, TimeUnit.MILLISECONDS)) {
			}
		} catch (Exception e) {
			logger.error("Error while calculating static cost table.", e);
		}
		logger.debug("--- end calculating static cost table ---");
	}

	/**
	 * updates the static cost table with new host switch and sorting costs
	 * 
	 * @param hostswithCosts
	 * @param sortingCosts
	 */
	public void update(BigDecimal hostswithCosts, BigDecimal sortingCosts) {
		for (StaticCostTableItem[] costItemArray : costs) {
			for (StaticCostTableItem costItem : costItemArray) {
				if (costItem.getCost() != null && costItem.isValid()) {
					costItem.setCost(BigDecimal.valueOf(costItem.getSortings())
							.multiply(sortingCosts).add(
									costItem.isHostswitch() ? hostswithCosts
											: BigDecimal.ZERO));
				}
			}
		}
	}

	/**
	 * @author Nicolas Wieseke
	 * 
	 */
	public class StaticCostCalculator implements Runnable {

		/**
		 * the start node
		 */
		private IHostNode start;
		/**
		 * the end node
		 */
		private IHostNode end;

		/*
		 * (non-Javadoc)
		 * 
		 * @see java.lang.Runnable#run()
		 */

		public void run() {
			int i = start.getNumber();
			int j = end.getNumber();

			if (NodeUtils.isChild(this.start, this.end)) {
				if (StaticCostTable.this.hostswithOption
						.equals(OptionsConstants.HOSTSWITCH_0_ID)) {
					calcHostswtichCosts();
				} else {
					StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
							i, j, new BigDecimal(Double.MAX_VALUE), 0, false,
							false);
				}
			} else if (i == j) {
				StaticCostTable.this.costs[i][j] = new StaticCostTableItem(i,
						j, BigDecimal.ZERO, 0, false, true);
			} else if (NodeUtils.isChild(this.end, this.start)) {
				int levelDifference = this.end.getLevel()
						- this.start.getLevel();
				levelDifference = Math.max(0, levelDifference - 1);
				StaticCostTable.this.costs[i][j] = new StaticCostTableItem(i,
						j, BigDecimal.valueOf(levelDifference).multiply(
								StaticCostTable.this.sortingCosts),
						levelDifference, false, true);
			} else {
				if (StaticCostTable.this.hostswithOption
						.equals(OptionsConstants.HOSTSWITCH_2_ID)) {
					StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
							i, j, new BigDecimal(Double.MAX_VALUE), 0, false,
							false);
				} else {
					calcHostswtichCosts();
				}
			}

		}

		/**
		 * sets the start and end node
		 * 
		 * @param start
		 * @param end
		 */
		public void init(IHostNode start, IHostNode end) {
			this.start = start;
			this.end = end;
		}

		/**
		 * calculates the costs in case a host switch occurred
		 */
		private void calcHostswtichCosts() {
			int i = start.getNumber();
			int j = end.getNumber();
			IHostNode endAncestor = this.end;
			int minLevelDistance = 0;
			List<Integer> startDescendantNumbers = new ArrayList<Integer>();
			List<Integer> endAncestorNumbers = new ArrayList<Integer>();
			List<IHostNode> endAncestors = new ArrayList<IHostNode>();

			// if time zones will be used
			if (StaticCostTable.this.timezoneOption
					.equals(OptionsConstants.RANK_1_ID)) {
				// if start time zone is greater that end time zone set infinity
				// cost
				if (checkTimezone(start, end) > 0) {
					StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
							i, j, new BigDecimal(Double.MAX_VALUE), 0, false,
							false);
					return;
				}

				startDescendantNumbers.add(this.start.getNumber());
				// find ancestor of node end with the right time zone
				IHostNode tempAncestor = this.end;
				while (tempAncestor.equals(this.end) || tempAncestor != null) {
					endAncestor = tempAncestor;
					endAncestors.add(0, tempAncestor);
					if (checkTimezone(this.start, tempAncestor) == 0)
						break;
					tempAncestor = tempAncestor.getParent();
				}
				endAncestorNumbers.add(endAncestor.getNumber());
				minLevelDistance = this.end.getLevel() - endAncestor.getLevel();
				// if the take-off can be delayed
				if (StaticCostTable.this.takeoffOption
						.equals(OptionsConstants.TAKEOFF_1_ID)) {
					// delay the take-off to find a host switch with fewer
					// sortings
					// the hostswitchCandidates contains all take-off and
					// associated landing sites
					Map<IHostNode, IHostNode> hostswitchCandidates = new LinkedHashMap<IHostNode, IHostNode>();
					hostswitchCandidates.put(this.start, endAncestor);
					// endAncestors contains all potential landing sites, i.e.
					// from the current endAncestor (the landing site of the
					// host switch with take-off site this.start, i.e. no
					// delayed take-off) to this.end
					for (int endAncestorIndex = 0; endAncestorIndex < endAncestors
							.size(); endAncestorIndex++) {
						IHostNode ancestor = endAncestors.get(endAncestorIndex);
						List<IHostNode> descendants = new ArrayList<IHostNode>();
						descendants.add(this.start);
						// to find a better take-off site for a fixed landing
						// site (ancestor) we check the descendants of
						// this.start which have a distance lower than
						// n, with n be the distance of the current ancestor
						// to the landing site belonging to this.start, i.e. we
						// take a potential landing site and search for all
						// possible descendants of this.start if there is a
						// valid take-off site which have a lower distance to
						// this.start than the distance between the potential
						// landing site and the landing site belonging to
						// this.start. Such a take-off and landing site pair
						// would result in a host switch with fewer sortings.
						for (int startDescendantIndex = 0; startDescendantIndex < endAncestorIndex; startDescendantIndex++) {
							// the list of next descendants are computed
							List<IHostNode> nextDescendants = new ArrayList<IHostNode>();
							for (IHostNode prevDescendant : descendants) {
								nextDescendants.addAll(prevDescendant
										.getChildren());
							}
							descendants = nextDescendants;
							// for each descendant the time zone validity is
							// checked concerning to the current landing site
							for (IHostNode descendant : descendants) {
								if (checkTimezone(descendant, ancestor) == 0) {
									int startLevelDifference = descendant
											.getLevel()
											- this.start.getLevel();
									int endLevelDifference = this.end
											.getLevel()
											- ancestor.getLevel();
									int levelDifference = startLevelDifference
											+ endLevelDifference;
									// if the sum of both distances
									// (levelDifference) is equal to
									// minLevelDistance than this
									// take-off-landing pair is another
									// candidate for our host switch, with the
									// same number of sortings
									if (minLevelDistance == levelDifference) {
										hostswitchCandidates.put(descendant,
												ancestor);
									}
									// if it is smaller then minLevelDistance
									// then this pair need fewer sortings and
									// therefore is our new and currently only
									// host switch candidate
									else if (minLevelDistance > levelDifference) {
										hostswitchCandidates.clear();
										hostswitchCandidates.put(descendant,
												ancestor);
										minLevelDistance = levelDifference;
									}
								}
							}
						}
					}

					startDescendantNumbers = new ArrayList<Integer>();
					endAncestorNumbers = new ArrayList<Integer>();
					// each host switch candidates have to be added to our list
					// of startDescendantNumbers and endAncestorNumbers
					for (Entry<IHostNode, IHostNode> entry : hostswitchCandidates
							.entrySet()) {
						startDescendantNumbers.add(entry.getKey().getNumber());
						endAncestorNumbers.add(entry.getValue().getNumber());
					}
				}
			}
			// if no time zones will be used the startDescendantNumbers
			// (respectively endAncestorNumbers) will only contain the
			// this.start (respectively this.end)
			else {
				startDescendantNumbers.add(this.start.getNumber());
				endAncestorNumbers.add(this.end.getNumber());
			}

			// no sorting costs
			if (StaticCostTable.this.sortingOption
					.equals(OptionsConstants.SORTING_0_ID)) {
				StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
						startDescendantNumbers, endAncestorNumbers,
						StaticCostTable.this.hostswithCosts, 0, true, true);
				return;
			}
			// costs for sortings along the rest of the path
			if (StaticCostTable.this.sortingOption
					.equals(OptionsConstants.SORTING_1_ID)) {
				// minLevelDistance = Math.max(0, minLevelDistance - 1);
				StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
						startDescendantNumbers, endAncestorNumbers,
						StaticCostTable.this.hostswithCosts
								.add(StaticCostTable.this.sortingCosts
										.multiply(BigDecimal
												.valueOf(minLevelDistance))),
						minLevelDistance, true, true);
				return;
			}
			// sorting costs for the whole path through the common ancestor
			if (StaticCostTable.this.sortingOption
					.equals(OptionsConstants.SORTING_2_ID)) {
				INode commonAncestor = NodeUtils.getCommonAncestor(this.start,
						this.end);
				int levelDifference = Math.abs(this.start.getLevel()
						- commonAncestor.getLevel())
						+ Math.abs(this.end.getLevel()
								- commonAncestor.getLevel());
				levelDifference = Math.max(0, levelDifference - 1);
				StaticCostTable.this.costs[i][j] = new StaticCostTableItem(
						startDescendantNumbers, endAncestorNumbers,
						StaticCostTable.this.hostswithCosts
								.add(StaticCostTable.this.sortingCosts
										.multiply(BigDecimal
												.valueOf(levelDifference))),
						levelDifference, true, true);
				return;
			}

			// this should never happen
			StaticCostTable.this.costs[i][j] = new StaticCostTableItem(i, j,
					new BigDecimal(Double.MAX_VALUE), 0, false, false);
		}

		/**
		 * @return -1 if the time zone of start is before time zone of end, 1 if
		 *         the time zone of start is after time zone of end and 0 if the
		 *         time zones overlap
		 */
		private int checkTimezone(IHostNode start, IHostNode end) {
			int startTimezoneFrom = start.getParent() == null ? Integer.MIN_VALUE
					: start.getParent().getRank();
			int startTimezoneTo = start.getRank();
			int endTimezoneFrom = end.getParent() == null ? Integer.MIN_VALUE
					: end.getParent().getRank();
			int endTimezoneTo = end.getRank();
			if (startTimezoneFrom > endTimezoneTo)
				return 1;
			if (startTimezoneTo < endTimezoneFrom)
				return -1;
			return 0;
		}
	}

	public String toCostString() {
		StringBuffer sb = new StringBuffer();
		for (StaticCostTableItem[] costsRow : this.costs) {
			for (StaticCostTableItem cost : costsRow) {
				StringBuffer sb2 = new StringBuffer();
				sb2.append(cost.getCost());
				sb2.setLength(10);
				sb.append(sb2.toString().replace('\u0000', ' '));
			}
			sb.append("\n");
		}
		return sb.toString();
	}

}
